# -*- coding: utf-8 -*-
#
# Copyright 2017-2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Click utilities."""

import click


class CaseInsensitiveChoice(click.Choice):
    """Case-insensitive click choice.

    Based on https://github.com/pallets/click/issues/569.
    """

    def convert(self, value, param, ctx):
        """Convert value to its choice value."""
        if value is None:
            return None
        return super(CaseInsensitiveChoice, self).convert(value.lower(), param, ctx)
